/*
 * Copyright 2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.myfaces.renderkit.html;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.faces.context.ResponseStream;
import javax.faces.context.ResponseWriter;
import javax.faces.context.FacesContext;
import javax.faces.render.RenderKit;
import javax.faces.render.Renderer;
import javax.faces.render.ResponseStateManager;
import java.io.OutputStream;
import java.io.Writer;
import java.io.IOException;
import java.util.*;


/**
 * @author Manfred Geiler (latest modification by $Author: baranda $)
 * @version $Revision: 278747 $ $Date: 2005-09-05 16:27:51 +0200 (Mon, 05 Sep 2005) $
 */
public class HtmlRenderKitImpl
    extends RenderKit
{
    private static final Log log = LogFactory.getLog(HtmlRenderKitImpl.class);
    private static String HTML_CONTENT_TYPE = "text/html";
    private static String FORM_SUBMIT_APPLICATION_WWW_FORM_URLENCODED_CONTENT_TYPE ="application/x-www-form-urlencoded";
    private static String FORM_SUBMIT_MULTIPART_FORM_DATA_CONTENT_TYPE="multipart/form-data";
    private static String ANY_CONTENT_TYPE ="*/*";

    //todo: enable these as soon as myfaces supports proper text/xml
    /*private static String XHTML_CONTENT_TYPE = "application/xhtml+xml";
    private static String APPLICATION_XML_CONTENT_TYPE = "application/xml";
    private static String TEXT_XML_CONTENT_TYPE = "text/xml";*/

    private static String DEFAULT_CHAR_ENCODING = "ISO-8859-1";

    //~ Instance fields ----------------------------------------------------------------------------

    private Map _renderers;
    private ResponseStateManager _responseStateManager;

    //~ Constructors -------------------------------------------------------------------------------

    public HtmlRenderKitImpl()
    {
        _renderers = new HashMap();
        _responseStateManager = new HtmlResponseStateManager();
    }

    //~ Methods ------------------------------------------------------------------------------------

    private String key(String componentFamily, String rendererType)
    {
        return componentFamily + "." + rendererType;
    }

    public Renderer getRenderer(String componentFamily, String rendererType)
    {
        if(componentFamily == null)
        {
            throw new NullPointerException("component family must not be null.");
        }
        if(rendererType == null)
        {
            throw new NullPointerException("renderer type must not be null.");
        }
        Renderer renderer = (Renderer) _renderers.get(key(componentFamily, rendererType));
        if (renderer == null)
        {
            log.warn("Unsupported component-family/renderer-type: " + componentFamily + "/" + rendererType);
        }
        return renderer;
    }

    public void addRenderer(String componentFamily, String rendererType, Renderer renderer)
    {
        if(componentFamily == null)
        {
            throw new NullPointerException("component family must not be null.");
        }
        if(rendererType == null)
        {
            throw new NullPointerException("renderer type must not be null.");
        }
        if(renderer == null)
        {
            throw new NullPointerException("renderer must not be null.");
        }
        _renderers.put(key(componentFamily, rendererType), renderer);
    }

    public ResponseStateManager getResponseStateManager()
    {
        return _responseStateManager;
    }

    public ResponseWriter createResponseWriter(Writer writer,
                                               String contentTypeListString,
                                               String characterEncoding)
    {
        FacesContext context = FacesContext.getCurrentInstance();

        if (contentTypeListString == null)
        {
            contentTypeListString = (String)
                    context.getExternalContext().getRequestHeaderMap().get("Accept");

            if(contentTypeListString == null)
            {
                if (log.isDebugEnabled())
                    log.debug("No content type list given, creating HtmlResponseWriterImpl with default content type.");

                contentTypeListString = HTML_CONTENT_TYPE;
            }
        }

        List contentTypeList = splitContentTypeListString(contentTypeListString);
        String[] supportedContentTypeArray = new String[]{HTML_CONTENT_TYPE,
                /*todo: enable these as soon as MyFaces supports proper text/xml markup
                XHTML_CONTENT_TYPE,APPLICATION_XML_CONTENT_TYPE,TEXT_XML_CONTENT_TYPE,*/
                FORM_SUBMIT_APPLICATION_WWW_FORM_URLENCODED_CONTENT_TYPE,
                FORM_SUBMIT_MULTIPART_FORM_DATA_CONTENT_TYPE,
                ANY_CONTENT_TYPE};

        String selectedContentType = null;

        for (int i = 0; i < contentTypeList.size(); i++)
        {
                String contentType = (String) contentTypeList.get(i);

                for (int j = 0; j < supportedContentTypeArray.length; j++)
                {
                    String supportedContentType = supportedContentTypeArray[j].trim();

                    if (contentType.indexOf(supportedContentType) != -1)
                    {
                        if (contentType.indexOf(HTML_CONTENT_TYPE) != -1 ||
                                contentType.indexOf(FORM_SUBMIT_APPLICATION_WWW_FORM_URLENCODED_CONTENT_TYPE) != -1 ||
                                contentType.indexOf(FORM_SUBMIT_MULTIPART_FORM_DATA_CONTENT_TYPE) != -1 ||
                                contentType.indexOf(ANY_CONTENT_TYPE)!=-1) {
                            selectedContentType = HTML_CONTENT_TYPE;
                        }
                        /* todo: enable this as soon as MyFaces supports proper text/xml markup
                        else if (contentType.indexOf(XHTML_CONTENT_TYPE) != -1 ||
                                 contentType.indexOf(APPLICATION_XML_CONTENT_TYPE) != -1 ||
                                 contentType.indexOf(TEXT_XML_CONTENT_TYPE) != -1) {
                            selectedContentType = XHTML_CONTENT_TYPE;
                        }*/
                        break;
                    }
                }
                if (selectedContentType!=null)
                {
                    break;
                }
	    }

        if(selectedContentType==null)
        {
            throw new IllegalArgumentException("ContentTypeList does not contain a supported content type: " +
                    contentTypeListString);
        }

        if(characterEncoding==null)
        {
            characterEncoding = DEFAULT_CHAR_ENCODING;
        }

        return new HtmlResponseWriterImpl(writer, selectedContentType, characterEncoding);
    }

    private List splitContentTypeListString(String contentTypeListString)
    {
        List contentTypeList = new ArrayList();

        StringTokenizer st = new StringTokenizer(contentTypeListString, ",");
        while (st.hasMoreTokens())
        {
            String contentType = st.nextToken().trim();

            int semicolonIndex = contentType.indexOf(";");

            if (semicolonIndex!=-1)
            {
                contentType = contentType.substring(0,semicolonIndex);
            }

            contentTypeList.add(contentType);
        }

        return contentTypeList;
    }

    public ResponseStream createResponseStream(OutputStream outputStream)
    {
        final OutputStream output = outputStream;

        return new ResponseStream()
        {
            public void write(int b) throws IOException
            {
                output.write(b);
            }


            public void write(byte b[]) throws IOException
            {
                output.write(b);
            }


            public void write(byte b[], int off, int len) throws IOException
            {
                output.write(b, off, len);
            }


            public void flush() throws IOException
            {
                output.flush();
            }


            public void close() throws IOException
            {
                output.close();
            }
        };
    }
}
